# import packages/modules
import os
import molinfo as mi

# check version
# print(mi.__version__)

# sdf file
# Methanol
# Conformer3D_COMPOUND_CID_887
# Benzene
# Conformer3D_COMPOUND_CID_241
# Glycol Dimethacrylate
# Conformer3D_COMPOUND_CID_7355
# 2-(2-Methoxyethoxy)ethanol
# Conformer3D_COMPOUND_CID_8134
# Bromobenzyl cyanide
# Conformer3D_COMPOUND_CID_22044
# Amineptine
# Conformer3D_COMPOUND_CID_34870
# Amidephrine
# Conformer3D_COMPOUND_CID_15010
# N,N-Dimethylformamide
# Conformer3D_COMPOUND_CID_6228
# Acetanilide
# Conformer3D_COMPOUND_CID_904
# Ascorbic Acid
# Conformer3D_COMPOUND_CID_54670067
# Malathion
# Conformer3D_COMPOUND_CID_4004
# Benzbromarone
# Structure2D_COMPOUND_CID_2333
# Butyraldehyde
# Structure2D_COMPOUND_CID_261
sdf_file_name_1 = 'test\Structure2D_COMPOUND_CID_261.sdf'
sdf_file = os.path.join(os.getcwd(), sdf_file_name_1)

# create graph
res = mi.create_graph(sdf_file)
print(type(res))
print(res)

# visualize compound by sdf file
# mi.g3d(sdf_file, display_bond_length=True)

# visualize compound by inchi
# mi.g3d_by_inchi(
#     'InChI=1S/C14H22O6/c1-11(2)13(15)19-9-7-17-5-6-18-8-10-20-14(16)12(3)4/h1,3,5-10H2,2,4H3')

# network
# res = mi.check_functional_group(sdf_file, functional_groups=['ether'])
# print(res)

# res = mi.check_functional_group(sdf_file, res_format='dataframe')
# print(res)

# res = mi.check_functional_group(sdf_file)
# print(res)



# NETWORK
# -------

# import packages/modules
import pandas as pd
import networkx as nx
from networkx.algorithms import isomorphism
# local
from .chemgraphs import ChemGraphs


class Network(ChemGraphs):

    def __init__(self, atomElements, atomBonds, xyzList, xyzCenterList, atomBonds1d):
        self.atomElements = atomElements
        # bond block (info)
        self.atomBonds = atomBonds
        self.xyzList = xyzList
        self.xyzCenterList = xyzCenterList

        # 1d vector of atom bonds
        self.atomBonds1d = atomBonds1d

        # TODO: super
        ChemGraphs.__init__(self)

        # functional group list
        self.function_group_list = {
            'hydroxyl': [self.graph_hydroxyl()],
            'carbonyl': [self.graph_carbonyl()],
            'carboxyl': [self.graph_carboxyl()],
            'ether': [self.graph_ether()],
            'pst-alcohols': [self.graph_primary_alcohol(), self.graph_secondary_alcohol(),
                             self.graph_tertiary_alcohol(), self.graph_secondary_alcohol_double_bond(),
                             self.graph_primary_alcohol_double_bond()],
            'primary-alcohol': [self.graph_primary_alcohol(), self.graph_primary_alcohol_double_bond()],
            'secondary-alcohol': [self.graph_secondary_alcohol(), self.graph_secondary_alcohol_double_bond()],
            'tertiary-alcohol': [self.graph_tertiary_alcohol()],
            'alkane': [self.graph_alkane()],
            'alkane CH bond (sp3)': [self.graph_alkane_CH_bond()],
            'alkene': [self.graph_alkene()],
            'alkene CH bond (sp2)': [self.graph_alkene_CH_bond()],
            'aromatic CH bond (sp2)': [self.graph_aromatic_CH_bond()],
            'alkyne': [self.graph_alkyne()],
            'alkyne CH bond (sp)': [self.graph_alkyne_CH_bond()],
            'arene': [self.graph_arene()],
            'aldehyde': [self.graph_aldehyde()],
            'ketone': [self.graph_ketone()],
            'carboxylic-acid': [self.graph_carboxylic_acid()],
            'ester': [self.graph_ester()],
            'pst-amide': [self.graph_primary_amide(), self.graph_secondary_amide(), self.graph_tertiary_amide()],
            'primary-amide': [self.graph_primary_amide()],
            'secondary-amide': [self.graph_secondary_amide()],
            'tertiary-amide': [self.graph_tertiary_amide()],
            'pst-amine': [self.graph_primary_amine(), self.graph_secondary_amine(), self.graph_tertiary_amine()],
            'primary-amine': [self.graph_primary_amine()],
            'secondary-amine': [self.graph_secondary_amine()],
            'tertiary-amine': [self.graph_tertiary_amine()],
            'nitrile': [self.graph_nitrile()],
            'thiol': [self.graph_thiol()],
            'alkyl-halids': [self.graph_alkyl_halide('F'), self.graph_alkyl_halide('Cl'), self.graph_alkyl_halide('Br'),
                             self.graph_alkyl_halide('I'),
                             self.graph_primary_alkyl_halide(
                                 'F'), self.graph_primary_alkyl_halide('Cl'),
                             self.graph_primary_alkyl_halide('Br'), self.graph_primary_alkyl_halide('I')]
        }

    def check_functional_groups(self, functional_groups=[]):
        '''
        Check functional groups in a compound

        Parameters
        ----------
        functional_groups : list
            list of functional groups

        Returns
        -------
        res : dict
            a list of all count
        '''
        # check functional group
        if len(functional_groups) == 0:
            functional_groups = list(self.function_group_list.keys())
        # create graph
        G = self.create_graph()
        # check functional groups
        res = self.check_functional_group(G, functional_groups)
        # res
        return res

    def check_functional_group(self, G, function_groups):
        '''
        Check a functional group exists in a compound

        Parameters
        ----------
        G : graph
            graph
        function_groups : list[str]
            functional group name like hydroxyl

        Returns
        -------
        res : dict
            a list of all count
        '''
        def node_match(n1, n2):
            '''
            Define a custom node_match function to ensure element matching
            '''
            return n1['symbol'] == n2['symbol']

        def edge_match(e1, e2):
            '''
            Define a custom edge_match function to ensure bond type matching
            '''
            return e1['type'] == e2['type']

        # res
        res_match = []

        # for each functional group
        for item in function_groups:
            # ! check functional exists in the list
            if item in self.function_group_list:
                # get a list of graphs
                function_group_graphs = self.function_group_list.get(item)

                # flag to track if functional group is found
                fg_found_any = False

                # graph function list
                for _fn in function_group_graphs:
                    # Create a GraphMatcher object for a functional group
                    fg_matcher = isomorphism.GraphMatcher(
                        G, _fn, node_match=node_match, edge_match=edge_match)

                    # Check if a functional group is in the main graph
                    fg_found = fg_matcher.subgraph_is_isomorphic()

                    if fg_found:
                        # print(f"{item} found in the molecule!")
                        fg_found_any = True
                        # res
                        res_match.append({
                            'function_group': item,
                            'result': fg_found
                        })
                        # break
                        break
                # check
                if not fg_found_any:
                    # print(f"{item} not found in the molecule.")
                    # res
                    res_match.append({
                        'function_group': item,
                        'result': False
                    })

        # res
        return res_match

    def create_graph(self):
        '''
        Check functional groups in a compound

        Parameters
        ----------
        functional_groups : list
            list of functional groups

        Returns
        -------
        res : dict
            a list of all count
        '''
        # atom no
        atomNo = len(self.xyzList)
        # bond no
        bondNo = len(self.atomBonds1d)

        # Create a graph from atoms and bonds
        G = nx.Graph()

        # *** atom visualization
        for i in range(atomNo):
            # xyz
            _atom1X = self.xyzList[i, 0]
            _atom1Y = self.xyzList[i, 1]
            _atom1Z = self.xyzList[i, 2]
            _atom1XYZ = [_atom1X, _atom1Y, _atom1Z]

            # color
            # atom id
            _atomId = int(i+1)
            # symbol
            _atomSymbol = str(self.atomElements[i]).strip()
            # atom mark
            atomMark = str(_atomSymbol) + str(_atomId)

            # add node with cooridination
            # G.add_node(_atomId, symbol=_atomSymbol,
            #            x=_atom1X, y=_atom1Y, z=_atom1Z, xyz=_atom1XYZ)

            G.add_node(_atomId, symbol=_atomSymbol)

        # *** using bond block

        for i in range(bondNo):
            # atom 1
            _id1 = self.atomBonds1d[i]['id1']
            _symbol1 = self.atomBonds1d[i]['symbol1']
            # atom 2
            _id2 = self.atomBonds1d[i]['id2']
            _symbol2 = self.atomBonds1d[i]['symbol2']
            # bond type
            _bondType = self.atomBonds1d[i]['bond_type']
            # bond symbol
            _bondSymbol = self.atomBonds1d[i]['bond_symbol']

            # add edge
            G.add_edge(_id1, _id2, symbol=_bondSymbol, type=_bondType)

        # res
        return G



# import packages/modules
import os
import pubchemquery as pcq
import pandas as pd
# internal
from .config import packageName
from .config import packageShortName
from .config import __version__
from .config import __description__
from .docs import MolParser
from .docs import Compound


def main():
    # short description and version
    _des = f"{packageShortName} {__version__}: {__description__}"
    print(_des)


def create_graph(file):
    '''
    Convert a sdf compound file to a graph 

    Parameters
    ----------
    file : str
        molecule file format (sdf)

    Returns
    -------
    graph : object
        compound graph
    '''
    try:
        if os.path.exists(file):
            # parse file
            MolParserC = MolParser(file)
            compound_info = MolParserC.read_file()
            # compound
            compound = Compound(compound_info)
            # display 3d
            graph = compound.create_graph()
            # res
            return graph
        else:
            raise Exception("file path is not valid.")
    except Exception as e:
        print(e)


def g3d(file, fig_size=[], bg_color='#ffffff', display_legend=True, display_atom_id=True, display_bond_length=False):
    '''
    3d graph of a compound

    Parameters
    ----------
    file : str
        molecule file format (sdf)
    display_legend : bool
        display legend (default True)
    display_atom_id : bool
        display atom id (default True)
    display_bond_length : bool
        display bond length (default True)

    Returns
    -------
    None
        display 3d graph
    '''
    # check file exists
    if os.path.exists(file):
        # parse file
        MolParserC = MolParser(file)
        compound_info = MolParserC.read_file()
        # compound
        compound = Compound(compound_info)
        # display 3d
        compound.view3d(fig_size=fig_size, display_legend=display_legend, bg_color=bg_color,
                        display_atom_id=display_atom_id, display_bond_length=display_bond_length)
    else:
        raise Exception("file path is not valid.")


def g3d_by_inchi(inchi, fig_size=[], bg_color='#ffffff', display_legend=True, display_atom_id=True, display_bond_length=False):
    '''
    3d graph of a compound using its InChI identifier

    Parameters
    ----------
    inchi : str
        inchi code
    display_legend : bool
        display legend (default True)

    Returns
    -------
    None
        display 3d graph
    '''
    # check inchi
    if inchi is not None:
        # get cid
        cid = pcq.get_cid_by_inchi(inchi)
        # check
        if cid is not None:
            # get sdf
            sdf = pcq.get_structure_by_cid(cid)
            # check
            if sdf is not None:
                # display 3d
                # parse file
                MolParserC = MolParser(None)
                compound_info = MolParserC.read_file(
                    sourceContent={
                        'content': sdf,
                        'format': 'sdf'
                    })
                # compound
                compound = Compound(compound_info)
                # display 3d
                compound.view3d(fig_size=fig_size, display_legend=display_legend, bg_color=bg_color,
                                display_atom_id=display_atom_id, display_bond_length=display_bond_length)
            else:
                raise Exception("sdf is not found!")
        else:
            raise Exception("cid is not found!")
    else:
        raise Exception("inchi is not valid.")


def check_functional_group(file, functional_groups=[], res_format='dict'):
    '''
    Check a functional group exists in a compound

    Parameters
    ----------
    file : str
        molecule file format (sdf)
    functional_groups : list[str]
        functional group (default ['hydroxyl'])
    res_format : str
        result format (default 'dict')

    Returns
    -------
    res : dict
        a list of all count
    '''
    # check file exists
    if os.path.exists(file):
        # parse file
        MolParserC = MolParser(file)
        compound_info = MolParserC.read_file()
        # compound
        compound = Compound(compound_info)
        # check functional group
        res = compound.check_functional_groups(functional_groups)
        # check
        if res_format == 'dict':
            return res
        elif res_format == 'dataframe':
            # dataframe
            return pd.DataFrame(res)
    else:
        raise Exception("file path is not valid.")


if __name__ == "__main__":
    main()



# MATERIAL PARSER
# ----------------

# import libs
import re
import numpy as np
import copy
# internals
from ..config import OBS_POSITIONS
from .structure import Structure
from .element import Element
from .utility import Utility


class MolParser():
    '''
    Parse different formats of molecule files such as sdf, json, ...
    '''
    _mat = {}

    def __init__(self, filepath):
        self.filepath = filepath

    @property
    def mat(self):
        return self._mat

    @mat.setter
    def mat(self, value):
        self._mat = copy.deepcopy(value)

    def read_file(self, sourceContent={}):
        '''
        Read structure file 
            1. from file
            2. download from api

        Parameters
        ----------
        filepath : str
            full file name with directory 

        Returns
        -------
        res : dict
            mol: mol object
            structure: structure object

        hints:
            mol: mol object
                header_block: header block
                counts_line: counts line
                atom_numbers: atom number
                mat_cid: cid number
                mat_name: IUPAC name
                mat_formula: formula
                mat_mass: molecular mass
                atom_names: atom list
                atom_xyz: atom coordinate
                bond_numbers: bond number
                bond_list: bond list
                bond_matrix: bond matrix
                bond_xyz: bond coordinate

            structure: structure object
                mol: mol object
                mat: mat object
        '''
        try:
            # get file path
            filePath = self.filepath

            # check
            if filePath:
                # open file and get content
                fileContent, fileDir, fileName, fileFormat = Utility.OpenFile(
                    filePath)
            else:
                # read string/json and ... files
                fileContent, fileFormat = Utility.ReadContent(
                    sourceContent)

            # method selection
            parserFun = {
                'sdf': self.sdf_parser,
                'json': self.json_parser
            }

            # parse file
            parserSelection = parserFun.get(fileFormat)
            parserRes = parserSelection(fileContent)
            return parserRes

        except Exception as e:
            raise Exception(e)

    def sdf_parser(self, sdfSource, sdfVersion='V2000'):
        '''
        Parse sdf file

        Parameters
        ----------
        sdfSource : str
            sdf file content

        sdfVersion : str
            sdf file version (default V2000)

        Returns
        -------
        res : dict
            mol: mol object
                header_block: header block
                counts_line: counts line
                atom_numbers: atom number
                mat_cid: cid number
                mat_name: IUPAC name
                mat_formula: formula
                mat_mass: molecular mass
                atom_names: atom list
                atom_elements: atom list
                bond_numbers: bond number
                atom_block: atoms
                bond_block: bond list
                xyz_list: xyz list
                xyz_center_list: xyz center list
                compound_properties: compound properties

        '''
        # decode binary
        # if binary file
        # sdfSource = sdfSource.decode('utf-8')
        # if text file
        sdfSource = sdfSource
        # create list
        sdfSourceList = sdfSource.splitlines()

        # header block
        headerBlock = sdfSourceList[0:3]

        # counts line
        countsLine = sdfSourceList[3]
        if countsLine.find('V2000') == -1:
            raise Exception(
                'SDF file version is not compatible with this method, import 2000 version.')
        # check
        countsLineList = countsLine.split()

        # find 'M END'
        MENDi = -1
        MENDi = sdfSourceList.index('M  END')
        # connection table
        if MENDi != -1:
            connectionTable = sdfSourceList[4:MENDi]

        # *** check
        _countsLineListSize = len(countsLineList)
        if _countsLineListSize == 10:
            # atom no
            atomNo = int(countsLineList[0])
            # bond no
            bondNo = int(countsLineList[1])
        elif _countsLineListSize == 9:
            _connectionTableLines = connectionTable
            _connectionTableLinesSize = len(_connectionTableLines)
            # first record
            _firstRecord = _connectionTableLines[0]
            _firstRecordSize = len(_firstRecord)
            _secondRecord = _connectionTableLines[0]
            _secondRecordSize = len(_secondRecord)
            if _firstRecordSize != _secondRecordSize:
                raise Exception('sdf file is not coded correctly.')
            # loop
            for l in range(_connectionTableLinesSize):
                _loopRecordSize = len(_connectionTableLines[l])
                if _loopRecordSize < _firstRecordSize:
                    # atom no
                    atomNo = int(l)
                    # bond no
                    bondNo = int(str(countsLineList[0]).split(str(atomNo))[1])
                    # break
                    break
        else:
            raise Exception('3rd line of the sdf file is not coded correctly.')

        # element rows
        elementRows = connectionTable[0:atomNo]
        # bond rows
        bondRows = connectionTable[atomNo:]

        # other vars
        compoundPropertiesList = MolParser.__var_finder(sdfSource)
        # dict vars
        compoundProperties = MolParser.__var_analyzer(compoundPropertiesList)
        # extract:
        PUBCHEM_COMPOUND_CID = compoundProperties.get('PUBCHEM_COMPOUND_CID')
        PUBCHEM_IUPAC_NAME = compoundProperties.get('PUBCHEM_IUPAC_NAME')
        PUBCHEM_EXACT_MASS = compoundProperties.get('PUBCHEM_EXACT_MASS')
        PUBCHEM_MOLECULAR_FORMULA = compoundProperties.get(
            'PUBCHEM_MOLECULAR_FORMULA')
        PUBCHEM_MOLECULAR_WEIGHT = compoundProperties.get(
            'PUBCHEM_MOLECULAR_WEIGHT')

        atoms = []
        atomList = []
        xyzList = []

        # atoms position
        for i in range(atomNo):
            _atomRow = elementRows[i].split()
            # position
            _x = float(_atomRow[0])
            _y = float(_atomRow[1])
            _z = float(_atomRow[2])
            # name
            _name = _atomRow[3]

            # atom list
            atomList.append(_name)

            # atom info
            atom = {
                'id': i+1,
                'symbol': _name,
                'index': i+1,
                'x': _x,
                'y': _y,
                'z': _z,
                'position': {
                    'x': _x,
                    'y': _y,
                    'z': _z
                },
                'xyz': [_x, _y, _z]
            }
            # save atom info
            atoms.append(atom)
            # xyz
            xyzList.append([_x, _y, _z])

        # set
        xyzList = np.array(xyzList)

        # object base
        objectBaseCoordinate = Structure.CenterPoints(xyzList)
        # print(f"objectBaseCoordinate: {objectBaseCoordinate}")

        # move to the center [0,0,0]
        xyzCenterList, movingCoordinate = Structure.CenterObject(
            xyzList, objectBaseCoordinate)
        # print(f"xyzCenterList: {xyzCenterList} \n movingCoordinate: \n {movingCoordinate}")

        # create mat formula
        matFormula = Structure.create_formula(atomList)
        # calculate molecular mass
        # matMass = 1

        # bond analysis
        bondList = []
        atomBonds = []

        # atomNo: *** atom id in the structure ***
        for i in range(atomNo):
            # name
            _nameAtom1 = str(atomList[i])
            # index [not duplicated element]
            # _nameAtom1Index = atomList.index(_nameAtom1)
            _nameAtom1Index = i+1
            # check bond type
            for j in range(bondNo):
                _bondRow = bondRows[j].split()
                # starts from 1 to ... (not 0)
                _nameAtomBondRow = int(_bondRow[0])
                # check bond exist
                if _nameAtom1Index == _nameAtomBondRow:
                    # atom 2 index
                    _indexAtom = int(_bondRow[1])
                    # atom 2 name
                    _nameAtom2 = str(atomList[_indexAtom-1])
                    # bound type
                    _bondType = int(_bondRow[2])
                    # str bond
                    _bondName = _nameAtom1 + _nameAtom2
                    # atom bond
                    atomBonds.append(
                        (_indexAtom, _nameAtom2, _bondName, _bondType))

            if len(atomBonds) > 0:
                # save
                _bondList = {
                    'id': _nameAtom1Index,
                    'symbol': _nameAtom1,
                    'bonds': atomBonds
                }
                bondList.append(_bondList)

            # reset
            atomBonds = []
            _bondList = {}

        # res
        res = {
            'header_block': headerBlock,
            'counts_line': countsLine,
            'atom_numbers': atomNo,
            'mat_cid': PUBCHEM_COMPOUND_CID,
            'mat_name': PUBCHEM_IUPAC_NAME if PUBCHEM_IUPAC_NAME is not None else '',
            'mat_formula': PUBCHEM_MOLECULAR_FORMULA if PUBCHEM_MOLECULAR_FORMULA is not None else matFormula,
            'mat_mass': PUBCHEM_EXACT_MASS if PUBCHEM_EXACT_MASS is not None else PUBCHEM_MOLECULAR_WEIGHT,
            'atom_names': atomList,
            'atom_elements': atomList,
            'bond_numbers': bondNo,
            'atom_block': atoms,
            'bond_block': bondList,
            'xyz_list': xyzList,
            'xyz_center_list': xyzCenterList,
            'compound_properties': compoundProperties
        }

        # return
        return res

    def json_parser(self, jsonSource, id_sort=True):
        '''
        parse json file

        Parameters
        ----------
        jsonSource: dict file
            json file content
        id_sort: bool
            if True, sort id

        Returns
        -------
        res: dict
        '''
        try:
            dictSource = jsonSource['PC_Compounds'][0]

            # analyze json file
            # *** id node
            if 'id' in dictSource:
                _idNode = dictSource['id']
                cid = self.__json_parser_id(_idNode)

            # *** atoms
            if 'atoms' in dictSource:
                _atomsNode = dictSource['atoms']
                # atom ids, element atomic number
                atomIds, _elementAtomicNumber = self.__json_parser_atoms(
                    _atomsNode)

            # atom numbers
            atomNo = len(atomIds)

            # atom list (element list)
            elementListRes, elementList = self.__find_element_symbol(
                _elementAtomicNumber)

            # *** bonds
            if 'bonds' in dictSource:
                _bonds = dictSource['bonds']
                # atom1 id, atom2 id, bond types
                atomId1, atomId2, bondTypes, bondMatrix = self.__json_parser_bonds(
                    _bonds)

            # bond numbers
            bondNo = len(atomId1)

            # *** coords
            if 'coords' in dictSource:
                _coords = dictSource['coords']  # list
                # coords type, aid, xyzList
                coords_type, coords_aid, xyzList, structure_type = self.__json_parser_coords(
                    _coords)

            # *** charge
            if 'charge' in dictSource:
                _charge = dictSource['charge']  # scaler
                charge = self.__json_parser_charge(_charge)

            # *** props
            if 'props' in dictSource:
                _props = dictSource['props']  # list of dict
                propDict = self.__json_parser_props(_props)

            # *** count
            if 'count' in dictSource:
                _count = dictSource['count']  # dict
                countList = self.__json_parser_count(_count)

            # interpret
            __json_atom_position_res = self.__json_atom_position(
                atomNo, elementList, xyzList)

            # bond block
            __json_atom_bondblock_res = self.__json_atom_bondblock(
                atomNo, bondNo, elementList, bondMatrix)

            # bondBlock: used for Matview class
            # atomDetails, bondBlock, xyzCenterList = self.__json_atom_analyzer(
            #     atomNo, bondNo, elementList, xyzList, bondMatrix)

            # *** origin info
            origin_info = {
                'atom_elements': __json_atom_position_res.get('elementList'),
                'atom_atomic_number': _elementAtomicNumber,
                'atom_details': __json_atom_position_res.get('atomDetails'),
                'bond_block': __json_atom_bondblock_res.get('bondBlock'),
                'bond_list': __json_atom_bondblock_res.get('bondMatrix'),
                'xyz_list': __json_atom_position_res.get('xyzList'),
                'xyz_center_list': __json_atom_position_res.get('xyzCenterList'),
            }

            # *** define new ids for mat and update
            # *** xyzList, bondMatrix, elementlist, elementAtomicNumber
            # *** bond_list = bondMatrix
            # *** atom_block = atom_details
            # if id_sort:
            mat_position_info, atom_id_conversion, xyz_list_sorted, \
                element_list_sorted, bond_list_sorted = self.SetAtomId(
                    xyzList, elementList, bondMatrix)

            # interpret
            __json_atom_position_res2 = self.__json_atom_position(
                atomNo, element_list_sorted, xyz_list_sorted)
            # set
            atom_block_sorted = __json_atom_position_res2.get('atomDetails')
            xyz_list_sorted = __json_atom_position_res2.get('xyzList')
            xyz_center_list_sorted = __json_atom_position_res2.get(
                'xyzCenterList')

            # bond block
            __json_atom_bondblock_res2 = self.__json_atom_bondblock(
                atomNo, bondNo, element_list_sorted, bond_list_sorted)
            # set
            bond_block_sorted = __json_atom_bondblock_res2.get('bondBlock')

            # atomic number sorted
            element_atomic_number = self.arrange_prop(
                _elementAtomicNumber, atom_id_conversion[:, 0])

            # update properties
            # name
            mat_name = propDict.get('IUPAC Name')
            if mat_name is None:
                mat_name = 'NULL'

            # formula
            mat_formula = propDict.get('Molecular Formula')
            if mat_formula is None:
                # create mat formula
                mat_formula = 'NULL'

            # molecular weight
            mat_mass = propDict.get('Molecular Weight')
            if mat_mass is None:
                # mat molecular weight/mass
                mat_mass = 0
            else:
                mat_mass = float(mat_mass)

            # res
            res = {
                'header_block': '',
                'counts_line': '',
                'mat_structure': structure_type,
                'mat_cid': cid,
                'mat_name': mat_name,
                'mat_formula': mat_formula,
                'mat_mass': mat_mass,
                'atom_numbers': atomNo,
                'atom_elements': element_list_sorted,
                'atom_atomic_number': element_atomic_number,
                'atom_block': atom_block_sorted,
                'bond_numbers': bondNo,
                'bond_block': bond_block_sorted,
                'bond_list': bond_list_sorted,
                'xyz_list': xyz_list_sorted,
                'xyz_center_list': xyz_center_list_sorted,
                'compound_properties': propDict,
                'mat_info_origin': origin_info
            }

            return res

        except Exception as e:
            raise Exception(e)

    def __json_parser_id(self, data):
        '''
        Return cid 

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        node : int
            cid
        '''
        # *** id node
        node = data['id']['cid']
        return node

    def __json_parser_atoms(self, data):
        '''
        Return atom id, atom atomic number

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        _aid : int
            atom id
        _elementAtomicNumber : int
            atom atomic number
        '''
        # aid (atom id)
        _aid = data['aid']
        # element (atomic number)
        _elementAtomicNumber = data['element']
        return _aid, _elementAtomicNumber

    def __json_parser_bonds(self, data):
        '''
        Return atom1 id, atom2 is, bond type

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        _aid1 : int
            atom1 id
        _aid2 : int
            atom2 id
        _order : int
            bond type
        '''
        # aid1 (atom id)
        _aid1 = data['aid1']
        # aid2 (atom id)
        _aid2 = data['aid2']
        # order
        _order = data['order']

        # transform
        bondMatrix = np.array([_aid1, _aid2, _order])
        bondMatrix = np.transpose(bondMatrix)

        return _aid1, _aid2, _order, bondMatrix

    def __json_parser_coords(self, data):
        '''
        Return coordination type, coordination id, xyzList

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        _coords_type : str
            coordination type
        _coords_aid : int
            coordination id
        xyzList : np.array
            xyzList
        structureType : str
            2d or 3d
        '''
        data = data[0]
        _coords_type = data['type']  # list
        _coords_aid = data['aid']  # list
        _coords_conformers = data['conformers'][0]
        _x = _coords_conformers.get('x')
        # atomNo
        atomNo = len(_x)
        _y = _coords_conformers.get('y')

        _z = _coords_conformers.get('z') if _coords_conformers.get(
            'z') is not None else np.zeros(atomNo)
        # transform
        xyzList = np.array([_x, _y, _z])
        xyzList = np.transpose(xyzList)

        # 2d/3d structure
        structureType = "2d" if np.count_nonzero(_z) == 0 else '3d'

        return _coords_type, _coords_aid, xyzList, structureType

    def __json_parser_charge(self, data):
        '''
        Get charge

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        node : int
            charge
        '''
        # *** id node
        node = data
        return node

    def __json_parser_props(self, data):
        '''
        Return a list of all properties

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        res : dict
            a list of all properties
        '''
        res = {}
        for i in range(len(data)):
            _keySet = data[i]['urn']['label']
            _valueSet = list(data[i]['value'].values())
            res[str(_keySet)] = _valueSet[0]

        return res

    def __json_parser_count(self, data):
        '''
        Return a list of all count

        Parameters
        ----------
        data : dict
            json data

        Returns
        -------
        res : dict
            a list of all count
        '''
        res = {}
        for key, value in data.items():
            res[key] = value
        return res

    def __json_atom_analyzer(self, atomNo, bondNo, elementList, xyzList, bondMatrix):
        '''
        Define xyz list and bond list

        Parameters
        ----------
        atomNo : int
            atom number
        bondNo : int
            bond number
        elementList : list
            element list
        xyzList : np.array
            xyz list
        bondMatrix : np.array
            bond matrix

        Returns
        -------
        atomDetails : list
            atom details
        objectBaseCoordinate : list
            object base coordinate
        bondDetails : list
            bond details
        '''
        # vars
        # *** detail about atoms
        atomDetails = []

        # element symbols
        atomList = elementList

        # atoms position
        for i in range(atomNo):
            _atomRow = xyzList[i, :]
            # position
            _x = float(_atomRow[0])
            _y = float(_atomRow[1])
            _z = float(_atomRow[2])
            # name
            _name = atomList[i]

            # atom info
            atom = {
                'id': i,
                'symbol': _name,
                'index': i,
                'x': _x,
                'y': _y,
                'z': _z,
                'position': {
                    'x': _x,
                    'y': _y,
                    'z': _z
                },
                'xyz': [_x, _y, _z]
            }
            # save atom info
            atomDetails.append(atom)

        # object base
        objectBaseCoordinate = Structure.CenterPoints(xyzList)
        # print(f"objectBaseCoordinate: {objectBaseCoordinate}")

        # move to the center [0,0,0]
        xyzCenterList, movingCoordinate = Structure.CenterObject(
            xyzList, objectBaseCoordinate)
        # print(f"xyzCenterList: {xyzCenterList} \n movingCoordinate: \n {movingCoordinate}")

        # bond analysis
        bondBlock = []
        atomBonds = []

        # atomNo: *** atom id in the structure ***
        for i in range(atomNo):
            # name
            _nameAtom1 = str(atomList[i])
            # index [not duplicated element]
            # _nameAtom1Index = atomList.index(_nameAtom1)
            _nameAtom1Index = i+1
            # check bond type
            for j in range(bondNo):
                _bondRow = bondMatrix[j, :]
                _nameAtomBondRow = int(_bondRow[0])
                # check bond exist
                if _nameAtom1Index == _nameAtomBondRow:
                    # atom 2 index
                    _indexAtom2 = int(_bondRow[1])
                    # atom 2 name
                    _nameAtom2 = str(atomList[_indexAtom2-1])
                    # bound type
                    _bondType = int(_bondRow[2])
                    # str bond
                    _bondName = _nameAtom1 + '-' + _nameAtom2
                    # str id bond
                    _bondId = str(_nameAtom1Index) + '-' + str(_indexAtom2)
                    # atom bond
                    atomBonds.append(
                        (_indexAtom2, _nameAtom2, _bondName, _bondType, _bondId))

            if len(atomBonds) > 0:
                # save
                _bondList = {
                    'id': _nameAtom1Index,
                    'symbol': _nameAtom1,
                    'bonds': atomBonds
                }
                bondBlock.append(_bondList)

            # reset
            atomBonds = []
            _bondList = {}

        # result
        return atomDetails, bondBlock, xyzCenterList

    def __json_atom_position(self, atomNo, elementList, xyzList):
        '''
        Set mat position into the center of cartesian coordination

        Parameters
        ----------
        atomNo : int
            atom number
        elementList : list
            element list
        xyzList : np.array
            xyz list

        Returns
        -------
        res : dict
            a list of all count
        '''
        try:
            # vars
            # *** detail about atoms
            atomDetails = []

            # element symbols
            atomList = elementList

            # atoms position
            for i in range(atomNo):
                _atomRow = xyzList[i, :]
                # position
                _x = float(_atomRow[0])
                _y = float(_atomRow[1])
                _z = float(_atomRow[2])
                # name
                _name = atomList[i]

                # atom info
                atom = {
                    'id': i,
                    'symbol': _name,
                    'index': i,
                    'x': _x,
                    'y': _y,
                    'z': _z,
                    'position': {
                        'x': _x,
                        'y': _y,
                        'z': _z
                    },
                    'xyz': [_x, _y, _z]
                }
                # save atom info
                atomDetails.append(atom)

            # object base
            objectBaseCoordinate = Structure.CenterPoints(xyzList)

            # move to the center [0,0,0]
            xyzCenterList, movingCoordinate = Structure.CenterObject(
                xyzList, objectBaseCoordinate)

            # res
            res = {
                'atomDetails': atomDetails,
                'elementList': elementList,
                'xyzList': xyzList,
                'xyzCenterList': xyzCenterList,
                'movingCoordinate': movingCoordinate
            }

            # res
            return res
        except Exception as e:
            raise Exception(e)

    def __json_atom_bondblock(self, atomNo, bondNo, elementList, bondMatrix):
        '''
        Build bond block which is used by mat view to display mat structure

        Parameters
        ----------
        atomNo : int
            atom number
        bondNo : int
            bond number
        elementList : list
            element list
        bondMatrix : np.array
            bond matrix

        Returns
        -------
        res : dict
            a list of all count
        '''
        try:
            # element symbols
            atomList = elementList

            # bond analysis
            bondBlock = []
            atomBonds = []

            # atomNo: *** atom id in the structure ***
            for i in range(atomNo):
                # name
                _nameAtom1 = str(atomList[i])
                # index [not duplicated element]
                _nameAtom1Index = int(i+1)
                # check bond type
                for j in range(bondNo):
                    # name
                    _bondRow = bondMatrix[j, :]
                    # atom id
                    _nameAtomBondRow = int(_bondRow[0])
                    # check bond exist
                    if _nameAtom1Index == _nameAtomBondRow:
                        # atom 2 index
                        _indexAtom2 = int(_bondRow[1])
                        # atom 2 name
                        _nameAtom2 = str(atomList[_indexAtom2-1])
                        # bound type
                        _bondType = int(_bondRow[2])
                        # str bond
                        _bondName = _nameAtom1 + '-' + _nameAtom2
                        # str id bond
                        _bondId = str(_nameAtom1Index) + '-' + str(_indexAtom2)
                        # atom bond
                        atomBonds.append(
                            (_indexAtom2, _nameAtom2, _bondName, _bondType, _bondId))

                if len(atomBonds) > 0:
                    # save
                    _bondList = {
                        'id': _nameAtom1Index,
                        'symbol': _nameAtom1,
                        'bonds': atomBonds
                    }
                    bondBlock.append(_bondList)

                # reset
                atomBonds = []
                _bondList = {}

            # res
            res = {
                'elementList': elementList,
                'bondMatrix': bondMatrix,
                'bondBlock': bondBlock
            }

            # res
            return res
        except Exception as e:
            Exception(e)

    def __find_element_symbol(self, atomic_numbers):
        '''
        Return element symbols

        Parameters
        ----------
        atomic_numbers : list
            atomic numbers

        Returns
        -------
        res : dict
            a list of all count
        atomList : list
            atom list
        '''
        el = Element()
        # loop
        res = el.find_atom_by_property('AtomicNumber', atomic_numbers)

        # interpret
        atomList = [item['symbol'] for item in res]

        # res
        return res, atomList

    def __var_finder(data):
        '''
        Find variables in a sdf file (single)

        Parameters
        ----------
        data : str
            sdf data

        Returns
        -------
        res : dict
            a list of all count
        '''
        res = re.findall(
            r"(\>\s*\<(.*)\s*\>\s*((.*\n)([^\<\>\$\$\$\$])*))", data, re.M)
        return res

    def __var_analyzer(data):
        '''
        Make a dict of all properties

        Parameters
        ----------
        data : str
            sdf data

        Returns
        -------
        res : dict
            a list of all count
        '''
        res = {}
        dataSize = len(data)
        for i in range(dataSize):
            # name
            varName = data[i][1]
            # val
            varVal = (data[i][2]).split('\n')
            varVal = [item.strip() for item in varVal]
            varVal = list(filter(None, varVal))

            # check
            if len(varVal) == 1:
                _varVal = str(varVal[0])
            else:
                _varVal = varVal
            # set
            _keySet = str(varName)
            _valueSet = _varVal
            # res
            res[_keySet] = _valueSet
            # reset

        return res

    def SetAtomId(self, xyzList, elementList, bondList):
        '''
        Set atom id with respect to their position in a 3d frame

        Parameters
        ----------
        xyzList : list
            list of element (atom) position
        elementList : list
            list of elements
        bondList : list
            list of bond ids and types

        Returns
        -------
        matPosition: list
            list of new ids
        xyzListSorted : list
            list of element (atom) position
        elementListSorted : list
            list of elements
        bondListSorted : list
            list of bond ids and types
        '''
        try:
            # robs position
            robs = OBS_POSITIONS
            # res
            disRes = []
            # calculate distance
            for i in range(len(xyzList)):
                _dis = np.linalg.norm(
                    np.array(robs) - np.array(xyzList[i]))
                _idOld = int(i+1)
                _symbol = elementList[i]
                disRes.append([_idOld, _dis, _symbol])

            # sort
            sortRes = sorted(disRes, key=lambda l: l[1], reverse=True)

            # reverse [old id, distance, symbol]
            sortRes.reverse()

            # save new id [old id, new id, distance, symbol]
            for j in range(len(sortRes)):
                _idNew = int(j+1)
                sortRes[j].insert(1, _idNew)

            # sorted id (old id, new id)
            idConversion = []
            for j in range(len(sortRes)):
                idConversion.append([sortRes[j][0], sortRes[j][1]])

            # build xyzList and elementlist with respect to the new ids
            xyzListSorted = []
            elementListSorted = []
            matPosition = {}
            for j in range(len(sortRes)):
                _idOld = sortRes[j][0]
                _idNew = sortRes[j][1]
                _dis = sortRes[j][2]
                _val1 = xyzList[_idOld-1]
                _val2 = elementList[_idOld-1]
                # set
                xyzListSorted.append(_val1)
                elementListSorted.append(_val2)
                # all
                matPosition[str(_idNew)] = [_idOld, _idNew, _dis, _val1, _val2]

            # set
            xyzListSorted = np.array(xyzListSorted)

            # bond id conversion
            bondList = np.array(bondList, dtype='i')
            # size
            bondRow = bondList.shape[0]
            # column 0
            bondColumn0 = np.array(bondList[:, 0])
            # column 1
            bondColumn1 = np.array(bondList[:, 1])
            # column 2
            bondColumn2 = np.array(bondList[:, 2])

            # replace
            for i in range(bondRow):
                # column 0
                _v0 = bondColumn0[i]
                # find new value
                _v1 = [item[1] for item in idConversion if item[0] == _v0][0]
                bondColumn0[i] = _v1

                # column 1
                _v2 = bondColumn1[i]
                # find new value
                _v3 = [item[1] for item in idConversion if item[0] == _v2][0]
                bondColumn1[i] = _v3

            # build bond list with new ids
            bondListSorted = np.zeros_like(bondList)
            bondListSorted[:, 0] = bondColumn0
            bondListSorted[:, 1] = bondColumn1
            bondListSorted[:, 2] = bondColumn2

            # set
            idConversion = np.array(idConversion)

            return matPosition, idConversion, xyzListSorted, elementListSorted, bondListSorted

        except Exception as e:
            Exception(e)

    def arrange_prop(self, property_value_list, atom_index_list):
        '''
        Arrange a new list with respect to the index list

        Parameters
        ----------
        property_value_list: list
            such as atomic number
        atom_index_list: list
            atom id between 1 and ...

        Returns
        -------
        res: list
            sorted list
        '''
        try:
            # convert atom id to list id
            index_list = np.array(atom_index_list) - 1
            property_value_list_sorted = np.take(
                property_value_list, index_list)
            # res
            return property_value_list_sorted
        except Exception as e:
            Exception(e)



# MATERIAL
# ----------

# import libs
import numpy as np
from matplotlib.pyplot import xlabel
from ..config import OBSERVER_PROPERTY
from .utility import Utility
from .graph3d import graph3d
from .netwrok import Network
from .compute import Compute, CalculateMolecularMass


class Compound(graph3d, Network):
    '''
    material along with properties
    '''
    _mat_cid = ''
    _mat_name = ''
    _mat_formula = ''
    _atom_numbers = 0
    _atom_elements = []
    _atom_xyz = []
    _atom_xyz_center = []
    _atom_bond_block = []
    _atom_bond_numbers = 0

    # *** obs fixed ***
    _limits = {
        'phi': [],
        'teta': [np.deg2rad(0), np.deg2rad(180)]
    }
    _robs = 10
    _tetaNo = OBSERVER_PROPERTY.get('TETA')
    _phiNo = OBSERVER_PROPERTY.get('PHI')
    _dataNo = []
    _obsCoordinate = []

    def __init__(self, parse_prop):
        # all properties
        self.parse_prop = parse_prop
        # super class
        __atom_elements = parse_prop['atom_elements']
        __atom_bonds = parse_prop['bond_block']
        __atom_xyz = parse_prop['xyz_list']
        __atom_xyz_center = parse_prop['xyz_center_list']
        # limit
        __limits = self._limits

        # TODO: convert atom_bonds to 1d vector
        __atom_bonds_1d = self.convert_atom_bonds(__atom_bonds)

        # ! init parent classes
        # *** raw info (just for visualizing a structure)
        graph3d.__init__(self, __atom_elements, __atom_bonds,
                         __atom_xyz, __atom_xyz_center, self.robs, self.tetaNo, self.phiNo, __limits, __atom_bonds_1d)

        # *** network
        Network.__init__(self, __atom_elements, __atom_bonds,
                         __atom_xyz, __atom_xyz_center, __atom_bonds_1d)

        # update mat prop
        self.__update_atom_prop('mat_cid')
        self.__update_atom_prop('mat_name')
        self.__update_atom_prop('mat_formula')
        self.__update_atom_prop('atom_numbers')
        self.__update_atom_prop('atom_elements')
        self.__update_atom_prop('bond_numbers')
        self.__update_atom_prop('bond_block')
        self.__update_atom_prop('xyz_list')
        self.__update_atom_prop('xyz_center_list')

    def __str__(self):
        '''
        Return info about the mat
        '''
        # check
        if self.parse_prop['mat_name'] is not None:
            return str(self.parse_prop['mat_name'])
        else:
            return str(self.parse_prop['atom_elements'])

    @property
    def mat_cid(self):
        return self.parse_prop['mat_cid']

    @mat_cid.setter
    def mat_cid(self, value):
        self._mat_cid = value

    @property
    def mat_name(self):
        return self.parse_prop['mat_name']

    @mat_name.setter
    def mat_name(self, value):
        self._mat_name = value

    @property
    def mat_formula(self):
        return self.parse_prop['mat_formula']

    @mat_formula.setter
    def mat_formula(self, value):
        self._mat_formula = value

    @property
    def atom_elements(self):
        return np.array(self.parse_prop['atom_elements'])

    @atom_elements.setter
    def atom_elements(self, value):
        self._atom_elements = value

    @property
    def atom_numbers(self):
        return self.parse_prop['atom_numbers']

    @atom_numbers.setter
    def atom_numbers(self, value):
        self._atom_numbers = value

    @property
    def atom_xyz(self):
        return np.array(self.parse_prop['xyz_list'])

    @atom_xyz.setter
    def atom_xyz(self, value):
        self._atom_xyz = value

    @property
    def atom_xyz_center(self):
        return np.array(self.parse_prop['xyz_center_list'])

    @atom_xyz_center.setter
    def atom_xyz_center(self, value):
        self._atom_xyz_center = value

    @property
    def atom_bond_block(self):
        return self.parse_prop['bond_block']

    @atom_bond_block.setter
    def atom_bond_block(self, value):
        self._atom_bond_block = value

    @property
    def atom_bond_numbers(self):
        return self.parse_prop['bond_numbers']

    @atom_bond_numbers.setter
    def atom_bond_numbers(self, value):
        self._atom_bond_numbers = value

    # *** observer prop
    @property
    def limits(self):
        return self._limits

    @limits.setter
    def limits(self, value):
        self._limits = []
        self._limits = value

    @property
    def robs(self):
        return self._robs

    @robs.setter
    def robs(self, value):
        self._robs = value

    @property
    def tetaNo(self):
        return self._tetaNo

    @tetaNo.setter
    def tetaNo(self, value):
        self._tetaNo = value

    @property
    def phiNo(self):
        return self._phiNo

    @phiNo.setter
    def phiNo(self, value):
        self._phiNo = value

    @property
    def dataNo(self):
        return self._dataNo

    @dataNo.setter
    def dataNo(self, value):
        self._dataNo = value

    @property
    def obsCoordinate(self):
        return self._obsCoordinate

    @obsCoordinate.setter
    def obsCoordinate(self, value):
        self._obsCoordinate = []
        self._obsCoordinate = value

    def __update_atom_prop(self, prop_name):
        '''
        Update atom prop

        Parameters
        ----------
        prop_name: str
            atom prop name
        '''
        switchProp = {
            'header_block': 1,
            'counts_line': 1,
            'mat_cid': self.__update_mat_cid,
            'mat_name': self.__update_mat_name,
            'mat_formula': self.__update_mat_formula,
            'atom_numbers': self.__update_atom_numbers,
            'atom_elements': self.__update_atom_elements,
            'bond_numbers': self.__update_atom_bond_numbers,
            'atom_block': 1,
            'bond_block': self.__update_atom_bond_block,
            'xyz_list': self.__update_atom_xyz,
            'xyz_center_list': self.__update_atom_xyz_center
        }
        # select prop
        propSelection = switchProp.get(prop_name)
        propSelection(self.parse_prop[prop_name])

    def __update_mat_cid(self, prop_val):
        self.mat_cid = prop_val

    def __update_mat_name(self, prop_val):
        self.mat_name = prop_val

    def __update_mat_formula(self, prop_val):
        self.mat_formula = prop_val

    def __update_atom_numbers(self, prop_val):
        self.atom_numbers = prop_val

    def __update_atom_elements(self, prop_val):
        self.atom_elements = prop_val

    def __update_atom_xyz(self, prop_val):
        self.atom_xyz = prop_val

    def __update_atom_xyz_center(self, prop_val):
        self.atom_xyz_center = prop_val

    def __update_atom_bond_block(self, prop_val):
        self.atom_bond_block = prop_val

    def __update_atom_bond_numbers(self, prop_val):
        self.atom_bond_numbers = prop_val

    def convert_atom_bonds(self, atom_bonds):
        '''
        Convert atom bonds to 1d list
        '''
        atom_bonds_1d = []
        for atom in atom_bonds:
            for bond in atom['bonds']:
                atom_bonds_1d.append({
                    'id1': atom['id'],
                    'symbol1': atom['symbol'],
                    'id2': bond[0],
                    'symbol2': bond[1],
                    'bond_type': bond[3],
                    'bond_symbol': bond[2]
                })
        return atom_bonds_1d

    def distance_matrix(self):
        '''
        Build a matrix of atom-atom distance
        '''
        return Compute.atoms_distance_matrix(self.xyzList, self.atom_elements)

    def distance_atoms(self, atom_symbols, atom_index=[]):
        '''
        Calculate distance between two different atoms

        Parameters
        ----------
        atom_symbols: list
            atom symbol list such as ['C','H']
        atom_index: list
            atom index such as [0,1]

        Returns
        -------
        distance: float
            distance between two atoms
        '''
        return Compute.atoms_distance(self.xyzList, self.atom_elements, atom_symbols, atom_index)



# DEFINE FUNCTIONAL GROUPS
# -----------------------

# import packages/modules
import networkx as nx


class ChemGraphs():

    def __init__(self):
        pass

    def graph_hydroxyl(self):
        '''
        Create a graph for hydroxyl
        '''
        # OH graph
        OH = nx.Graph()
        # O
        OH.add_node(1, symbol='O')
        # H
        OH.add_node(2, symbol='H')
        # bond
        OH.add_edge(1, 2, symbol='OH', type=1)

        # res
        return OH

    def graph_carbonyl(self):
        '''
        Create a graph for carbonyl
        '''
        # C=O graph
        CO = nx.Graph()
        # C
        CO.add_node(1, symbol='C')
        # O
        CO.add_node(2, symbol='O')
        # bond
        CO.add_edge(1, 2, symbol='CO', type=2)

        # res
        return CO

    def graph_carboxyl(self):
        '''
        Create a graph for carboxyl
        '''
        # C=O graph
        CO = nx.Graph()
        # C
        CO.add_node(1, symbol='C')
        # O
        CO.add_node(2, symbol='O')
        # O
        CO.add_node(3, symbol='O')
        # H
        CO.add_node(4, symbol='H')
        # bond
        CO.add_edge(1, 2, symbol='CO', type=2)
        # bond
        CO.add_edge(1, 3, symbol='CO', type=1)
        # bond
        CO.add_edge(3, 4, symbol='OH', type=1)

        # res
        return CO

    def graph_ether(self):
        '''
        Create a graph for ether
        '''
        # COC graph
        COC = nx.Graph()
        # C
        COC.add_node(1, symbol='C')
        # O
        COC.add_node(2, symbol='O')
        # C
        COC.add_node(3, symbol='C')
        # bond
        COC.add_edge(1, 2, symbol='CO', type=1)
        # bond
        COC.add_edge(2, 3, symbol='OC', type=1)

        # res
        return COC

    def graph_alkane(self):
        '''
        Create a graph for alkane
        '''
        # C-C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=1)

        # res
        return CC

    def graph_alkene(self):
        '''
        Create a graph for alkene
        '''
        # C=C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=2)

        # res
        return CC

    def graph_alkyne(self):
        '''
        Create a graph for alkyne
        '''
        # C≡C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=3)

        # res
        return CC

    def graph_arene(self):
        '''
        Create a graph for arene
        '''
        # C6H5 graph (simplified)
        C6H5 = nx.Graph()
        # C
        C6H5.add_node(1, symbol='C')
        # C
        C6H5.add_node(2, symbol='C')
        # C
        C6H5.add_node(3, symbol='C')
        # C
        C6H5.add_node(4, symbol='C')
        # C
        C6H5.add_node(5, symbol='C')
        # C
        C6H5.add_node(6, symbol='C')
        # bond
        C6H5.add_edge(1, 2, symbol='CC', type=1)
        # bond
        C6H5.add_edge(2, 3, symbol='CC', type=2)
        # bond
        C6H5.add_edge(3, 4, symbol='CC', type=1)
        # bond
        C6H5.add_edge(4, 5, symbol='CC', type=2)
        # bond
        C6H5.add_edge(5, 6, symbol='CC', type=1)
        # bond
        C6H5.add_edge(6, 1, symbol='CC', type=2)

        # res
        return C6H5

    def graph_aldehyde(self):
        '''
        Create a graph for aldehyde
        '''
        # CHO graph
        CHO = nx.Graph()
        # C
        CHO.add_node(1, symbol='C')
        # H
        CHO.add_node(2, symbol='H')
        # O
        CHO.add_node(3, symbol='O')
        # C
        CHO.add_node(4, symbol='C')
        # bond
        CHO.add_edge(1, 2, symbol='CH', type=1)
        # bond
        CHO.add_edge(1, 3, symbol='CO', type=2)
        # bond
        CHO.add_edge(1, 4, symbol='CC', type=1)

        # res
        return CHO

    def graph_ketone(self):
        '''
        Create a graph for ketone
        '''
        # COC graph
        COC = nx.Graph()
        # C
        COC.add_node(1, symbol='C')
        # O
        COC.add_node(2, symbol='O')
        # C
        COC.add_node(3, symbol='C')
        # C
        COC.add_node(4, symbol='C')
        # bond
        COC.add_edge(1, 2, symbol='CO', type=2)
        # bond
        COC.add_edge(1, 3, symbol='CC', type=1)
        # bond
        COC.add_edge(1, 4, symbol='CC', type=1)

        # res
        return COC

    def graph_carboxylic_acid(self):
        '''
        Create a graph for carboxylic acid
        '''
        # COOH graph
        COOH = nx.Graph()
        # C
        COOH.add_node(1, symbol='C')
        # O
        COOH.add_node(2, symbol='O')
        # O
        COOH.add_node(3, symbol='O')
        # H
        COOH.add_node(4, symbol='H')
        # C
        COOH.add_node(5, symbol='C')
        # bond
        COOH.add_edge(1, 2, symbol='CO', type=2)
        # bond
        COOH.add_edge(1, 3, symbol='CO', type=1)
        # bond
        COOH.add_edge(3, 4, symbol='OH', type=1)
        # bond
        COOH.add_edge(1, 5, symbol='CC', type=1)

        # res
        return COOH

    def graph_ester(self):
        '''
        Create a graph for ester
        '''
        # COOC graph
        COOC = nx.Graph()
        # C
        COOC.add_node(1, symbol='C')
        # O
        COOC.add_node(2, symbol='O')
        # O
        COOC.add_node(3, symbol='O')
        # C
        COOC.add_node(4, symbol='C')
        # C
        COOC.add_node(5, symbol='C')
        # bond
        COOC.add_edge(1, 2, symbol='CO', type=2)
        # bond
        COOC.add_edge(1, 3, symbol='CO', type=1)
        # bond
        COOC.add_edge(3, 4, symbol='OC', type=1)
        # bond
        COOC.add_edge(1, 5, symbol='CC', type=1)

        # res
        return COOC

    def graph_amide(self):
        '''
        Create a graph for amide
        '''
        # REVIEW
        # CONH2 graph
        CONH2 = nx.Graph()
        # C
        CONH2.add_node(1, symbol='C')
        # O
        CONH2.add_node(2, symbol='O')
        # N
        CONH2.add_node(3, symbol='N')
        # H
        CONH2.add_node(4, symbol='H')
        # H
        CONH2.add_node(5, symbol='H')
        # C
        CONH2.add_node(6, symbol='C')
        # bond
        CONH2.add_edge(1, 2, symbol='CO', type=2)
        # bond
        CONH2.add_edge(1, 3, symbol='CN', type=1)
        # bond
        CONH2.add_edge(3, 4, symbol='NH', type=1)
        # bond
        CONH2.add_edge(3, 5, symbol='NH', type=1)
        # bond
        CONH2.add_edge(1, 6, symbol='CC', type=1)

        # res
        return CONH2

    def graph_primary_amide(self):
        '''
        Create a graph for a primary amide
        '''
        # CONH2 graph
        CONH2 = nx.Graph()
        # C
        CONH2.add_node(1, symbol='C')
        # O
        CONH2.add_node(2, symbol='O')
        # N
        CONH2.add_node(3, symbol='N')
        # H
        CONH2.add_node(4, symbol='H')
        # H
        CONH2.add_node(5, symbol='H')
        # C
        CONH2.add_node(6, symbol='C')
        # bond
        CONH2.add_edge(1, 2, symbol='CO', type=2)
        # bond
        CONH2.add_edge(1, 3, symbol='CN', type=1)
        # bond
        CONH2.add_edge(3, 4, symbol='NH', type=1)
        # bond
        CONH2.add_edge(3, 5, symbol='NH', type=1)
        # bond
        CONH2.add_edge(1, 6, symbol='CC', type=1)

        # res
        return CONH2

    def graph_secondary_amide(self):
        '''
        Create a graph for a secondary amide
        '''
        # CONH2 graph
        CONH2 = nx.Graph()
        # C
        CONH2.add_node(1, symbol='C')
        # O
        CONH2.add_node(2, symbol='O')
        # N
        CONH2.add_node(3, symbol='N')
        # H
        CONH2.add_node(4, symbol='H')
        # H
        CONH2.add_node(5, symbol='C')
        # C
        CONH2.add_node(6, symbol='C')
        # bond
        CONH2.add_edge(1, 2, symbol='CO', type=2)
        # bond
        CONH2.add_edge(1, 3, symbol='CN', type=1)
        # bond
        CONH2.add_edge(3, 4, symbol='NH', type=1)
        # bond
        CONH2.add_edge(3, 5, symbol='NC', type=1)
        # bond
        CONH2.add_edge(1, 6, symbol='CC', type=1)

        # res
        return CONH2

    def graph_tertiary_amide(self):
        '''
        Create a graph for a tertiary amide
        '''
        # CONH2 graph
        CONH2 = nx.Graph()
        # C
        CONH2.add_node(1, symbol='C')
        # O
        CONH2.add_node(2, symbol='O')
        # N
        CONH2.add_node(3, symbol='N')
        # H
        CONH2.add_node(4, symbol='C')
        # H
        CONH2.add_node(5, symbol='C')
        # C
        CONH2.add_node(6, symbol='C')
        # bond
        CONH2.add_edge(1, 2, symbol='CO', type=2)
        # bond
        CONH2.add_edge(1, 3, symbol='CN', type=1)
        # bond
        CONH2.add_edge(3, 4, symbol='NC', type=1)
        # bond
        CONH2.add_edge(3, 5, symbol='NC', type=1)
        # bond
        CONH2.add_edge(1, 6, symbol='CC', type=1)

        # res
        return CONH2

    def graph_primary_amine(self):
        '''
        Create a graph for primary amine
        '''
        # NH2 graph
        NH2 = nx.Graph()
        # N
        NH2.add_node(1, symbol='N')
        # H
        NH2.add_node(2, symbol='H')
        # H
        NH2.add_node(3, symbol='H')
        # C
        NH2.add_node(4, symbol='C')
        # bond
        NH2.add_edge(1, 2, symbol='NH', type=1)
        # bond
        NH2.add_edge(1, 3, symbol='NH', type=1)
        # bond
        NH2.add_edge(1, 4, symbol='NC', type=1)

        # res
        return NH2

    def graph_secondary_amine(self):
        '''
        Create a graph for secondary amine
        '''
        # NH graph
        NH = nx.Graph()
        # N
        NH.add_node(1, symbol='N')
        # H
        NH.add_node(2, symbol='H')
        # C
        NH.add_node(3, symbol='C')
        # C
        NH.add_node(4, symbol='C')
        # bond
        NH.add_edge(1, 2, symbol='NH', type=1)
        # bond
        NH.add_edge(1, 3, symbol='NC', type=1)
        # bond
        NH.add_edge(1, 4, symbol='NC', type=1)

        # res
        return NH

    def graph_tertiary_amine(self):
        '''
        Create a graph for tertiary amine
        '''
        # N graph
        N = nx.Graph()
        # N
        N.add_node(1, symbol='N')
        # C
        N.add_node(2, symbol='C')
        # C
        N.add_node(3, symbol='C')
        # C
        N.add_node(4, symbol='C')
        # bond
        N.add_edge(1, 2, symbol='NC', type=1)
        # bond
        N.add_edge(1, 3, symbol='NC', type=1)
        # bond
        N.add_edge(1, 4, symbol='NC', type=1)

        # res
        return N

    def graph_nitrile(self):
        '''
        Create a graph for nitrile
        '''
        # CN graph
        CN = nx.Graph()
        # C
        CN.add_node(1, symbol='C')
        # N
        CN.add_node(2, symbol='N')
        # C
        CN.add_node(3, symbol='C')
        # bond
        CN.add_edge(1, 2, symbol='CN', type=3)
        # bond
        CN.add_edge(1, 3, symbol='NC', type=1)

        # res
        return CN

    def graph_thiol(self):
        '''
        Create a graph for thiol
        '''
        # SH graph
        SH = nx.Graph()
        # S
        SH.add_node(1, symbol='S')
        # H
        SH.add_node(2, symbol='H')
        # C
        SH.add_node(3, symbol='C')
        # bond
        SH.add_edge(1, 2, symbol='SH', type=1)
        # bond
        SH.add_edge(1, 3, symbol='SC', type=1)

        # res
        return SH

    def graph_alkyl_halide(self, halid):
        '''
        Create a graph for primary alkyl halide
        '''
        # R-X graph (simplified)
        RX = nx.Graph()
        # C
        RX.add_node(1, symbol='C')
        # X (halogen)
        RX.add_node(2, symbol=f'{halid}')
        # bond
        RX.add_edge(1, 2, symbol=f'C{halid}', type=1)

        # res
        return RX

    def graph_primary_alkyl_halide(self, halid):
        '''
        Create a graph for primary alkyl halide
        '''
        # R-X graph (simplified)
        RX = nx.Graph()
        # C
        RX.add_node(1, symbol='C')
        # C
        RX.add_node(2, symbol='C')
        # H
        RX.add_node(3, symbol='H')
        # H
        RX.add_node(4, symbol='H')
        # H
        RX.add_node(5, symbol='H')
        # X (halogen)
        RX.add_node(6, symbol=f'{halid}')
        # bond
        RX.add_edge(1, 2, symbol='CC', type=1)
        # bond
        RX.add_edge(1, 3, symbol='CH', type=1)
        # bond
        RX.add_edge(1, 4, symbol='CH', type=1)
        # bond
        RX.add_edge(1, 5, symbol='CH', type=1)
        # bond
        RX.add_edge(2, 6, symbol=f'C{halid}', type=1)

        # res
        return RX

    def graph_secondary_alkyl_halide(self):
        '''
        Create a graph for secondary alkyl halide
        '''
        # R-X graph (simplified)
        RX = nx.Graph()
        # C
        RX.add_node(1, symbol='C')
        # C
        RX.add_node(2, symbol='C')
        # C
        RX.add_node(3, symbol='C')
        # H
        RX.add_node(4, symbol='H')
        # H
        RX.add_node(5, symbol='H')
        # X (halogen)
        RX.add_node(6, symbol='X')
        # bond
        RX.add_edge(1, 2, symbol='CC', type=1)
        # bond
        RX.add_edge(1, 3, symbol='CC', type=1)
        # bond
        RX.add_edge(1, 4, symbol='CH', type=1)
        # bond
        RX.add_edge(1, 5, symbol='CH', type=1)
        # bond
        RX.add_edge(2, 6, symbol='CX', type=1)

        # res
        return RX

    def graph_tertiary_alkyl_halide(self):
        '''
        Create a graph for tertiary alkyl halide
        '''
        # R-X graph (simplified)
        RX = nx.Graph()
        # C
        RX.add_node(1, symbol='C')
        # C
        RX.add_node(2, symbol='C')
        # C
        RX.add_node(3, symbol='C')
        # C
        RX.add_node(4, symbol='C')
        # X (halogen)
        RX.add_node(5, symbol='X')
        # bond
        RX.add_edge(1, 2, symbol='CC', type=1)
        # bond
        RX.add_edge(1, 3, symbol='CC', type=1)
        # bond
        RX.add_edge(1, 4, symbol='CC', type=1)
        # bond
        RX.add_edge(1, 5, symbol='CX', type=1)

        # res
        return RX

    def graph_primary_alcohol(self):
        '''
        Create a graph for primary alcohol
        '''
        # R-OH graph (simplified)
        ROH = nx.Graph()
        # C
        ROH.add_node(1, symbol='C')
        # H
        ROH.add_node(2, symbol='H')
        # H
        ROH.add_node(3, symbol='H')
        # O
        ROH.add_node(4, symbol='O')
        # H
        ROH.add_node(5, symbol='H')
        # bond
        ROH.add_edge(1, 2, symbol='CH', type=1)
        # bond
        ROH.add_edge(1, 3, symbol='CH', type=1)
        # bond
        ROH.add_edge(1, 4, symbol='CO', type=1)
        # bond
        ROH.add_edge(4, 5, symbol='OH', type=1)

        # res
        return ROH

    def graph_primary_alcohol_double_bond(self):
        '''
        Create a graph for primary alcohol
        '''
        # R-OH graph (simplified)
        ROH = nx.Graph()
        # C
        ROH.add_node(1, symbol='C')
        # H
        ROH.add_node(2, symbol='H')
        # H
        ROH.add_node(3, symbol='C')
        # O
        ROH.add_node(4, symbol='O')
        # H
        ROH.add_node(5, symbol='H')
        # bond
        ROH.add_edge(1, 2, symbol='CH', type=1)
        # bond
        ROH.add_edge(1, 3, symbol='CC', type=2)
        # bond
        ROH.add_edge(1, 4, symbol='CO', type=1)
        # bond
        ROH.add_edge(4, 5, symbol='OH', type=1)

        # res
        return ROH

    def graph_secondary_alcohol(self):
        '''
        Create a graph for secondary alcohol
        '''
        # R-OH graph (simplified)
        ROH = nx.Graph()
        # C
        ROH.add_node(1, symbol='C')
        # C
        ROH.add_node(2, symbol='C')
        # C
        ROH.add_node(3, symbol='C')
        # H
        ROH.add_node(4, symbol='H')
        # O
        ROH.add_node(5, symbol='O')
        # H
        ROH.add_node(6, symbol='H')
        # bond
        ROH.add_edge(1, 2, symbol='CC', type=1)
        # bond
        ROH.add_edge(1, 3, symbol='CC', type=1)
        # bond
        ROH.add_edge(1, 4, symbol='CH', type=1)
        # bond
        ROH.add_edge(1, 5, symbol='CO', type=1)
        # bond
        ROH.add_edge(5, 6, symbol='OH', type=1)

        # res
        return ROH

    def graph_secondary_alcohol_double_bond(self):
        '''
        Create a graph for secondary alcohol
        '''
        # R-OH graph (simplified)
        ROH = nx.Graph()
        # C
        ROH.add_node(1, symbol='C')
        # C
        ROH.add_node(2, symbol='C')
        # C
        ROH.add_node(3, symbol='C')
        # O
        ROH.add_node(4, symbol='O')
        # H
        ROH.add_node(5, symbol='H')
        # bond
        ROH.add_edge(1, 2, symbol='CC', type=1)
        # bond
        ROH.add_edge(1, 3, symbol='CC', type=2)
        # bond
        ROH.add_edge(1, 4, symbol='CO', type=1)
        # bond
        ROH.add_edge(4, 5, symbol='OH', type=1)

        # res
        return ROH

    def graph_tertiary_alcohol(self):
        '''
        Create a graph for tertiary alcohol
        '''
        # R-OH graph (simplified)
        ROH = nx.Graph()
        # C
        ROH.add_node(1, symbol='C')
        # C
        ROH.add_node(2, symbol='C')
        # C
        ROH.add_node(3, symbol='C')
        # C
        ROH.add_node(4, symbol='C')
        # O
        ROH.add_node(5, symbol='O')
        # H
        ROH.add_node(6, symbol='H')
        # bond
        ROH.add_edge(1, 2, symbol='CC', type=1)
        # bond
        ROH.add_edge(1, 3, symbol='CC', type=1)
        # bond
        ROH.add_edge(1, 4, symbol='CC', type=1)
        # bond
        ROH.add_edge(1, 5, symbol='CO', type=1)
        # bond
        ROH.add_edge(5, 6, symbol='OH', type=1)

        # res
        return ROH

    def graph_alkane_CH_bond(self):
        '''
        Create a graph for alkane CH bond
        '''
        # C-C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # H
        CC.add_node(3, symbol='H')
        # H
        CC.add_node(4, symbol='H')
        # H
        CC.add_node(5, symbol='H')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=1)
        # bond
        CC.add_edge(1, 3, symbol='CH', type=1)
        # bond
        CC.add_edge(1, 4, symbol='CH', type=1)
        # bond
        CC.add_edge(1, 5, symbol='CH', type=1)

        # res
        return CC

    def graph_alkene_CH_bond(self):
        '''
        Create a graph for alkene
        '''
        # C=C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # H
        CC.add_node(3, symbol='H')
        # H
        CC.add_node(4, symbol='H')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=2)
        # bond
        CC.add_edge(1, 3, symbol='CH', type=1)
        # bond
        CC.add_edge(1, 4, symbol='CH', type=1)

        # res
        return CC

    def graph_aromatic_CH_bond(self):
        '''
        Create a graph for alkene
        '''
        # C=C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # C
        CC.add_node(3, symbol='C')
        # H
        CC.add_node(4, symbol='H')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=2)
        # bond
        CC.add_edge(1, 3, symbol='CC', type=1)
        # bond
        CC.add_edge(1, 4, symbol='CH', type=1)

        # res
        return CC

    def graph_alkyne_CH_bond(self):
        '''
        Create a graph for alkyne
        '''
        # C≡C graph
        CC = nx.Graph()
        # C
        CC.add_node(1, symbol='C')
        # C
        CC.add_node(2, symbol='C')
        # H
        CC.add_node(3, symbol='H')
        # bond
        CC.add_edge(1, 2, symbol='CC', type=3)
        # bond
        CC.add_edge(1, 3, symbol='CH', type=1)

        # res
        return CC




# COMPUTATIONAL
# ---------------

# import libs
import numpy as np


class Compute():
    '''
    computational chemistry
    '''

    def __init__(self):
        pass

    @staticmethod
    def cal_atoms_distance(xyzAtom1, xyzAtom2):
        '''
        calculate distance between two atoms (center to center)
        '''
        return np.linalg.norm(xyzAtom1-xyzAtom2)

    @staticmethod
    def atoms_distance_matrix(xyzList, atomName):
        '''
        build a matrix containing a matrix of distance between two different atoms

        args:
            xyzList: xyz list of atoms
            atomName: atom name list such as ['C','H','H','H','H']
        '''
        # atom no
        atomNo = len(atomName)
        atomLength = np.zeros((atomNo, atomNo))
        for i in range(atomNo):
            # atom1
            atom1XYZ = xyzList[i]
            for j in range(atomNo):
                # atom2
                atom2XYZ = xyzList[j]
                if j != i:
                    _length = Compute.cal_atoms_distance(atom1XYZ, atom2XYZ)
                else:
                    _length = 0
                # save
                atomLength[i, j] = _length

        # res
        return atomLength

    @staticmethod
    def atoms_distance(xyzList, atomName, atom_symbols, atom_index):
        '''
        build a matrix containing a matrix of distance between two different atoms

        args:
            xyzList: xyz list of atoms
            atomName: atom name list such as ['C','H','H','H','H']
        '''
        try:
            # size
            atom_symbols_size = len(atom_symbols)
            atom_index_size = len(atom_index)

            # check
            if atom_symbols_size > 2 or atom_index_size > 2:
                raise Exception(
                    'atom symbol/index list must have two elements.')

            if atom_index_size == 0:
                # find the index of the first element
                atom1Index = np.where(atomName == atom_symbols[0])[0][0]
                atom2Index = np.where(atomName == atom_symbols[1])[0][0]
                # set
                atom1XYZ = xyzList[atom1Index]
                atom2XYZ = xyzList[atom2Index]
                _length = Compute.cal_atoms_distance(atom1XYZ, atom2XYZ)
            elif atom_index_size == 2:
                # set
                atom1XYZ = xyzList[atom_index[0]]
                atom2XYZ = xyzList[atom_index[1]]
                _length = Compute.cal_atoms_distance(atom1XYZ, atom2XYZ)
            elif atom_index_size > 2:
                raise Exception(
                    'atom symbol/index list must have two elements.')

            # res
            return _length
        except Exception as e:
            raise


# UTILITY FUNCTION
# ------------------


def CalculateMolecularMass(atom_elements, element_source):
    '''
    calculate molecular mass [g/mol]

    args:
        atom_elements: such as C, H
        element_source: periodic element table
    '''
    try:
        # # check
        # if element_source is None:
        #     # load periodic table of elements
        #     element_source = elements()
        # res
        res = 0
        # atom element size
        atomElementsSize = len(atom_elements)
        # atom property
        _atom_property = ['AtomicMass']
        # loop
        for i in range(atomElementsSize):
            _atom_symbol = atom_elements[i]
            _atom_prop_res = element_source.atom_properties(
                _atom_symbol, _atom_property)
            _atom_prop_val = _atom_prop_res.get(_atom_property[0])
            res += float(_atom_prop_val)

        return res
    except Exception as e:
        raise



from .app import main, __version__, g3d, g3d_by_inchi, check_functional_group, create_graph

__all__ = ['main', '__version__', 'g3d',
           'g3d_by_inchi', 'check_functional_group', 'create_graph']
